# mpASTC

Multiprocessor, multi-node ASTC compressor

mpASTC is a parallel texture compressor for the Adaptive Scalable Texture
Compression (ASTC) format.

The goal of mpASTC is to produce high-quality compressed textures in real-time
through parallelism.

## Folder Structure

`mpastc/`:        The mpASTC library  
`mpastc/codegen`: Code generator and automatically-generated header files  
`app/`:           Encoding application  

## Building

You need the following packages to build mpASTC:

- C & C++ compiler supporting C11 and C++11, with OpenMP support
- CMake 3.9.0+
- GNU make 4.0+
- An MPI 3.0+ implementation (app only, if MPI support is enabled)

Check your distribution or cluster configuration for where these can be found.

You must generate partition and encoding combination code **before** compiling
mpASTC. For size reasons this is not included in the repository.

Go to the folder `mpastc/codegen` and run `make`. This will generate the header
files used by the mpASTC library. Verify the header files `enccomb.h` and
`partition.h` have been created.

If you want to compile the mpASTC **user-facing** application, you can do this
after code-generation by changing into the `app/` directory. Simply run the
command `cmake .` \- you can define the following as well with the format
`-DOPTION=1`:

- `ENABLE_MPI`: Enable MPI work distribution support.
- `ENABLE_DEBUG`: Enable debug asserts and flags
- `ENABLE_SAN`: Enable address & undefined behaviour sanitizers

Then run `make`. While CMake might claim the variables are unused, you will see output such as 
`MPI support enabled` if you did this correctly.

Short help can be requested from the application with `-h` or `--help`.

Images must be 3-channel JPG or PNG images (i.e., no purely b/w images or images with transparency).

### Sample Commands

Compress image  
`./mpaenc --file kodim01.jpg --output kodim01.astc`  
Produces the ASTC image with the ARM ASTC encoder header- you can then do  
`astcenc-native -dl kodim01.astc kodim01.png`  
Note you need the ARM ASTC encoder installed for this.

Compress image in test mode  
`./mpaenc --file kodim01.jpg --output kodim01.png -d`  
Immediately decodes the intermediate calculations instead of creating ASTC blocks.  
You can immediately view this image in your photo viewer without decoding.

**Warning**: You must specify when you are using MPI by using the `-m` flag.
If you do not pass the flag `-m` to mpaenc and you use MPI, all processes will
perform the same task and there will be no inter-rank parallelization! Likewise, if you are **not** using MPI and pass the flag `-m`, the application will not work correctly!

### Library

The `libmpastc` library is designed in cases where you have raw texture data ready to compress in real-time in your own program.
Should you wish to use it, you only need
to compile the program in the `mpastc/` folder- also with `cmake . && make`. Either compile and
copy the `.a` file and headers into your own program, or, if you use CMake in your own program as well, add
the `mpastc` folder as a subdirectory and link against `mpastc`.


## License

Copyright (c) 2023 Tim Dettmar

mpASTC incorporates the following third-party software:

- Headers from the [ARM reference ASTC encoder](https://github.com/ARM-software/astc-encoder), Apache 2.0 License
- log.c logging library: [link](https://github.com/rxi/log.c), MIT License
- stb_image from the stb libraries: [link](https://github.com/nothings/stb), MIT License

mpASTC is licensed under the BSD 3-Clause License. See the file LICENSE for
details.
